"""DEPRECATED, use resqpy.weights_and_measures instead"""

import warnings
from resqpy.weights_and_measures import *  # noqa

warnings.warn(
   "resqpy.olio.weights_and_measures is Deprecated and will be removed. Please update your code to use resqpy.weights_and_measures"
)
